{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TERATERM.EXE, 'Abort printing' dialogbox}
unit TeraPrn;

interface
{$I teraterm.inc}

{$IFDEF Delphi}
uses Messages, WinTypes, WinProcs, OWindows, ODialogs, Strings, CommDlg, TTTypes;
{$ELSE}
uses WinTypes, WinProcs, WObjects, Strings, CommDlg, Win31, TTTypes;
{$ENDIF}

const

IdPrnAbortDlg = 100;

type

  PPrnAbortDlg = ^TPrnAbortDlg;
  TPrnAbortDlg = object(TDlgWindow)
    procedure WMInitDialog(var Msg: TMessage);
      virtual wm_First + wm_InitDialog;
    procedure WMCommand(var Msg: TMessage);
      virtual wm_First + wm_Command;
    procedure CloseWindow; virtual;
  end;

var
  PrnDlg: TPrintDlg;
  PrintDC: HDC;
  Printing: boolean;
  PrintAbort: bool;
  PrintEX: bool;

  function PrnBox(HWin: HWnd; Sel: boolean): BOOL;
  function PrnStart(AParent: PWindowsObject; DocumentName: PChar): boolean;
  procedure PrnStop;

{printer emulation}
const
  PrnFName = 'PRINT.DAT';  
  IdPrnNormal = 1;
  IdPrnESC = 2;
  IdPrnCSI = 3;
  IdPrnCSIi = 4;
var
  AutoPrintMode: bool;
  PrinterMode: bool;
  HPrnFile: integer;
  PrnParseMode: integer;
  DB_CSI: bool;
  PCM: byte;

  procedure OpenPrnFile;
  procedure ClosePrnFile;
  procedure PrnParse(b: byte);

implementation

var
  PrnAbortDlg: PPrnAbortDlg;
  HPrnAbortDlg: HWnd;
  PPrnAbortProc: TFarProc;

{Print Abortion Call Back Function}
function PrnAbortProc(PDC: HDC; Code: integer): bool; export;
var
  Msg: TMsg;
begin
  while (not PrintAbort) and PeekMessage(Msg, 0,0,0, pm_Remove) do
  if (HPrnAbortDlg=0) or (not IsDialogMessage(HPrnAbortDlg, Msg)) then
  begin
    TranslateMessage(Msg);
    DispatchMessage(Msg);
  end;

  if PrintAbort then
  begin
    if PrnAbortDlg <> nil then PrnAbortDlg^.CloseWindow;
    PrnAbortDlg := nil;
  end;

  PrnAbortProc := not PrintAbort;
end;

function PrnBox(HWin: HWnd; Sel: boolean): BOOL;
begin
  PrnBox := false;
  {initialize TPrintDlg record}
  FillChar(PrnDlg, SizeOf(TPrintDlg), #0);
  with PrnDlg do
  begin
    lStructSize := SizeOf(TPrintDlg);
    hWndOwner := HWin;
    Flags := PD_ReturnDC or PD_NoPageNums or PD_ShowHelp;
    if not Sel then
      Flags := Flags or PD_NoSelection;
    nCopies := 1;
  end;
  {'Print' dialog box}
  if not PrintDlg(PrnDlg) then exit; {if 'Cancel' button, exit}
  if PrnDlg.hDC = 0 then exit;
  PrintDC := PrnDlg.hDC;
  PrnBox := true;
end;


function PrnStart(AParent: PWindowsObject; DocumentName: PChar): boolean;
var
  DocInfo: TDocInfo;
  DocName: array[0..31] of char;
begin
  PrnStart := false;
  Printing := false;
  PrintAbort := false;

  PrnAbortDlg := PPrnAbortDlg(Application^.MakeWindow(New(PPrnAbortDlg,
                      Init(AParent, PChar(IdPrnAbortDlg)))));
  if PrnAbortDlg = nil then exit;
  HPrnAbortDlg := PrnAbortDlg^.HWindow;
  PPrnAbortProc := MakeProcInstance(@PrnAbortProc, HInstance);
  SetAbortProc(PrintDC,TAbortProc(PPrnAbortProc));

  DocInfo.cbSize := SizeOf(TDocInfo);
  StrLCopy(DocName,DocumentName,31);
  DocInfo.lpszDocName := DocName;
  DocInfo.lpszOutput := nil;
  if StartDoc(PrintDC, DocInfo) > 0 then
  begin
    PrnStart := true;
    Printing := true;
  end
  else
    if PrnAbortDlg <> nil then PrnAbortDlg^.CloseWindow;
    
end;

procedure PrnStop;
begin
  if Printing then
  begin
    EndDoc(PrintDC);
    DeleteDC(PrintDC);
    Printing := false;
  end;
  if PrnAbortDlg <> nil then
  begin
    PrnAbortDlg^.CloseWindow;
    PrnAbortDlg := nil;
  end;
end;

{print abort dialog}
procedure TPrnAbortDlg.WMInitDialog(var Msg: TMessage);
begin
  SetFocus(HWindow);
end;

procedure TPrnAbortDlg.WMCommand(var Msg: TMessage);
begin
  PrintAbort := true;
end;

procedure TPrnAbortDlg.CloseWindow;
begin
  SetFocus(Parent^.HWindow);
  TDlgWindow.CloseWindow;
end;

{printer emulation routines}
procedure OpenPrnFile;
begin
  if HPrnFile > 0 then exit;
  HPrnFile := _lopen(PrnFName,OF_WRITE);
  if HPrnFile <= 0 then
    HPrnFile := _lcreat(PrnFName,0);
  if HPrnFile > 0 then
    _llseek(HPrnFile,0,2);
end;

procedure ClosePrnFile;
begin
  if HPrnFile > 0 then
    _lclose(HPrnFile);
  HPrnFile := 0;
end;

procedure WriteToPrnFile(b: byte);
begin
  case b of
    NUL: ;
    DC1: ;
    DC3: ;
  else
    _lwrite(HPrnFile,@b,1);
  end;
end;

procedure PrnParse(b: byte);
begin
  if HPrnFile<=0 then
  begin
    PrinterMode := false;
    exit;
  end;

  case PrnParseMode of
    IdPrnNormal:
      case b of
        ESC: PrnParseMode := IdPrnEsc;
        CSI: begin
           DB_CSI := false;
           PrnParseMode := IdPrnCSI;
         end;
      else
        WriteToPrnFile(b);
      end;
    IdPrnEsc:
      begin
        if b=Ord('[') then
        begin
          DB_CSI := true;
          PrnParseMode := IdPrnCSI;
        end
        else begin
          WriteToPrnFile(ESC);
          PrnParseMode := IdPrnNormal;
          PrnParse(b);
        end;
      end;
    IdPrnCSI:
      if (b=$34) or (b=$35) then
      begin
        PCM := b;
        PrnParseMode := IdPrnCSIi;
      end
      else begin
        if DB_CSI then
        begin
          WriteToPrnFile(ESC);
          WriteToPrnFile(Ord('['));
        end
        else
          WriteToPrnFile(CSI);
        PrnParseMode := IdPrnNormal;
        PrnParse(b);
      end;
    IdPrnCSIi:
      if b=Ord('i') then
      begin        
        PrinterMode := PCM=$35;
        if not PrinterMode and
           not AutoPrintMode then
          ClosePrnFile;
        PrnParseMode := IdPrnNormal;
      end
      else begin
        if DB_CSI then
        begin
          WriteToPrnFile(ESC);
          WriteToPrnFile(Ord('['));
        end
        else
          WriteToPrnFile(CSI);
        WriteToPrnFile(PCM);
        PrnParseMode := IdPrnNormal;
        PrnParse(b);
      end;
  else
      PrnParseMode := IdPrnNormal;
      PrnParse(b);
  end;
end;

begin
  AutoPrintMode := false;
  PrinterMode := false;
  HPrnFile := 0;
  PrintEX := true;
end.