{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TTFILE.DLL, file transfer, VT window printing}
library TTFile;
{$R ttfile.res}
{$I teraterm.inc}

{$IFDEF Delphi}
uses Messages, WinTypes, WinProcs, CommDlg, Strings, WinDos, ShellAPI,
     TTTypes, TTFTypes, DlgLib, Kermit, XMODEM,
     ZMODEM, BPlus, QuickVAN, VTPrn, TTLib, FTLib;
{$ELSE}
uses WinTypes, WinProcs, Win31, CommDlg, Strings, WinDos, ShellAPI,
     TTTypes, TTFTypes, DlgLib, Kermit, XMODEM,
     ZMODEM, BPlus, QuickVAN, VTPrn, TTLib, FTLib;
{$ENDIF}

const
  Id_GetFnDlg = 2000;
  Id_GetFn    = 2001;
  Id_GetFnHelp = 2099;

const
  Id_FnDlg = 2100;
  Id_FnBox = 2101;
  Id_FnXCheck = 2102;
  Id_FnX1K = 2104;
  ID_FnBinary = 2105;
  ID_FnAppend = 2106;
  Id_FnDlg2 = 2200;
  Id_Fn2Binary = 2201;

function GetSetupFname(HWin: HWnd; Save: BOOL; ts: PTTSet): BOOL; export;
var
  i, j: integer;
  OpenFN: TOpenFileName;
  Ptr: integer;
  FNameFilter: array [0..80] of char;
  TempDir: array[0..fsDirectory] of char;
  Dir: array[0..fsDirectory] of char;
  Name: array[0..fsPathName] of char;
  Ok: bool;
begin 
  {save current dir}
  GetCurDir(TempDir,0);

  {File name filter}
  FillChar(FNameFilter, SizeOf(FNameFilter), #0);
  StrCopy(FNameFilter, 'setup files(*.ini)');
  Ptr := StrLen(FNameFilter) + 1;
  StrCopy(@FNameFilter[Ptr], '*.ini'); 

  {TOpenFileName record}
  FillChar(OpenFN, SizeOf(TOpenFileName), #0);
  with OpenFN do
  begin
    lStructSize   := sizeof(TOpenFileName);
    hwndOwner     := HWin;
    lpstrFile     := Name;
    nMaxFile      := SizeOf(Name);
    lpstrFilter   := FNameFilter;
    lpstrDefExt   := 'ini';
    lpTemplateName := PChar(Id_FnDlg);
  end;
  OpenFN.hInstance := hinstance;

  GetFileNamePosJ(ts^.SetupFName,i,j);
  StrCopy(Name,@ts^.SetupFName[j]);
  Move(ts^.SetupFName[0],Dir[0],i);
  Dir[i] := #0;

  if (StrLen(Name)=0) or (StrIComp(Name,'TERATERM.INI')=0) then
  begin
    StrCopy(Name,'TERATERM.INI');
    OpenFN.nFilterIndex := 1;
  end
  else
    OpenFN.nFilterIndex := 2;

  if StrLen(Dir)=0 then
    StrCopy(Dir,ts^.HomeDir);

  SetCurDir(Dir);

  OpenFN.flags := ofn_ShowHelp or ofn_HideReadOnly or ofn_EnableTemplate;
  if Save then
  begin
    OpenFN.lpstrTitle := 'Tera Term: Save setup';
    Ok := GetSaveFileName(OpenFN);
  end
  else begin
    OpenFN.flags := OpenFN.flags or ofn_FileMustExist;
    OpenFN.lpstrTitle := 'Tera Term: Restore setup';
    Ok := GetOpenFileName(OpenFN);
  end;

  if Ok then
    StrCopy(ts^.SetupFName,Name);

  {restore dir}
  SetCurDir(TempDir);

  GetSetupFname := Ok;
end; 

function TFn2Hook(Dialog: HWnd; Message, WParam: Word; LParam: Longint): Bool; export;
var
  ofn: POpenFileName;
  w: word;
begin
  case Message of
    wm_InitDialog:
      begin
        ofn := POpenFileName(lParam);
        SetWindowLong(Dialog, DWL_USER, lParam);
        if ofn^.lCustData=0 then w := 0
                            else w := 1;
        ShowDlgItem(Dialog,Id_Fn2Binary,Id_Fn2Binary);
        SetRB(Dialog,w,Id_Fn2Binary,Id_Fn2Binary);
        TFn2Hook := true;
        exit;
      end;
    wm_Command:
      case wParam of
        id_Ok: begin
          ofn := POpenFileName(GetWindowLong(Dialog,DWL_USER));
          if ofn<>nil then
          begin
            GetRB(Dialog,w,Id_Fn2Binary,Id_Fn2Binary);
            ofn^.lCustData := w;
          end;
        end;
        id_Cancel: ;
      end;
  end;
  TFn2Hook := False;
end;

function GetMultiFname(HWin: HWnd; FileVar: PFileVar; CurDir: PChar; FuncId: word; Option: PWord): bool; export;
var
  i: integer;
  FNFilter: array [0..10] of Char;
  FNameRec: TOpenFileName;
  TempDir: array[0..fsDirectory] of char;
  Ok: BOOL;
begin
with FileVar^ do begin
  GetMultiFname := false;
  {save current dir}
  GetCurDir(TempDir,0);

  NumFname := 0;

  StrCopy(FileVar^.DlgCaption,'Tera Term: ');
  case FuncId of
    GMF_Kermit: StrCat(FileVar^.DlgCaption,'Kermit Send');
    GMF_Z: StrCat(FileVar^.DlgCaption,'ZMODEM Send');
    GMF_QV: StrCat(FileVar^.DlgCaption,'Quick-VAN Send');
  else
    exit;
  end;

  FnStrMemHandle := GlobalAlloc(gmem_Moveable, FnStrMemSize);
  if FnStrMemHandle = 0 then
  begin
    MessageBeep(0);
    exit;
  end
  else begin
    FnStrMem := GlobalLock(FnStrMemHandle);
    if FnStrMem = nil
    then begin
      GlobalFree(FnStrMemHandle);
      FnStrMemHandle := 0;
      MessageBeep(0);
      exit;
    end
  end;

  PChar(FnStrMem)[0] := #0;

  FillChar(FNFilter, SizeOf(FNFilter), #0);  { Set up for double null at end }
  StrCopy(FNFilter, 'all');
  StrCopy(@FNFilter[StrLen(FNFilter)+1], '*.*');
  FillChar(FNameRec, SizeOf(TOpenFileName), #0);
  with FNameRec do
  begin
    lStructSize   := sizeof(TOpenFileName);
    hwndOwner     := HWin;
    lpstrFilter   := FNFilter;
    nFilterIndex := 1;
    lpstrFile := FnStrMem;
    nMaxFile := FnStrMemSize;
    lpstrTitle:= FileVar^.DlgCaption;
    lpstrInitialDir := CurDir;
    flags := ofn_ShowHelp or ofn_AllowMultiSelect or
             ofn_FileMustExist or ofn_HideReadOnly or ofn_EnableTemplate;
    if FuncId=GMF_Z then
    begin
      flags := flags or ofn_EnableHook;
      lCustData := Option^;
      @lpfnHook := MakeProcInstance(@TFn2Hook, HInstance);
    end;
    lpTemplateName := PChar(Id_FnDlg2);
  end;
  FNameRec.hInstance := hinstance;

  Ok := GetOpenFileName(FNameRec);

  if Ok then
  begin
    if FuncId=GMF_Z then
      Option^ :=  FNameRec.lCustData and 1;

    {count number of file names}
    i := 0;
    while PChar(FnStrMem)[i]<>#0 do
    begin
      if PChar(FnStrMem)[i]=' ' then
      begin
        inc(NumFname);
        if NumFname = 1 then DirLen := i+1;
      end;
      inc(i);
    end;

    if NumFname<1 then
    begin
      NumFname := 1;
      DirLen := FNameRec.nFileOffset;
      StrCopy(FullName,PChar(FnStrMem));
    end
    else begin
      Move(PChar(FnStrMem)[0],FullName[0],DirLen-1);
      FullName[DirLen] := #0;
    end;
    FullName[DirLen-1] := '\';

    Move(FileVar^.FullName[0],CurDir[0],FileVar^.DirLen-1);
    CurDir[FileVar^.DirLen-1] := #0;

    FNcount := 0;
    if NumFname=1 then FNPtr := 0
                  else FNPtr := DirLen;
  end;

  GlobalUnlock(FnStrMemHandle);
  if not Ok then
  begin
    GlobalFree(FnStrMemHandle);
    FnStrMemHandle := 0;
  end;

  {restore dir}
  SetCurDir(TempDir);

  GetMultiFname := Ok;
end;
end;

function GetFnDlg(Dialog: HWnd; Message, WParam: Word;
  LParam: Longint): Bool; export;
var
  fv: PFileVar;
  TempFull: array[0..fsPathName] of char;
  i, j: integer;
begin
  GetFnDlg := True;
  case Message of
    wm_InitDialog:
      begin
        fv := PFileVar(lParam);
        SetWindowLong(Dialog, DWL_USER, lParam);
        SendDlgItemMessage(Dialog, id_GetFn, em_LimitText, SizeOf(TempFull)-1,0);
        Exit;
      end;
    wm_Command:
      begin
        fv := PFileVar(GetWindowLong(Dialog,DWL_USER));
        case wParam of
          id_Ok: begin
            if fv<>nil then
            begin
              with fv^ do begin
                GetDlgItemText(Dialog, id_GetFn, TempFull, SizeOf(TempFull));
                if StrLen(TempFull)=0 then exit;
                GetFileNamePosJ(TempFull,i,j);
                FitFileName(@TempFull[j],nil);
                StrCat(fv^.FullName,@TempFull[j]);
              end;
            end;
            EndDialog(Dialog, 1);
            exit;
          end;
          id_Cancel: begin
            EndDialog(Dialog, 0);
            exit;
          end;
          id_GetFnHelp:
            if fv<>nil then            
              PostMessage(fv^.HMainWin,WM_User+MsgDlgHelp2,0,0);
        end;
      end;
  end;
  GetFnDlg := False;
end;

function GetGetFname(HWin: HWnd; FileVar: PFileVar): Bool; export;
var
  GetFnProc: TFarProc;
begin
  GetFnProc := MakeProcInstance(@GetFnDlg, HInstance);
  GetGetFname := Bool( DialogBoxParam(HInstance,PChar(Id_GetFnDlg), HWin, GetFnProc, longint(FileVar)) );
  FreeProcInstance(GetFnProc);
  exit;
end;

procedure SetFileVar(FileVar: PFileVar); export;
var
  i: integer;
begin
with FileVar^ do begin
  GetFileNamePosJ(FullName,DirLen,i);
  if FullName[DirLen]='\' then inc(DirLen);
  StrCopy(FileVar^.DlgCaption,'Tera Term: ');
  case OpId of
    OpLog:      StrCat(DlgCaption,TitLog);
    OpSendFile: StrCat(DlgCaption,TitSendFile);
    OpKmtRcv:   StrCat(DlgCaption,TitKmtRcv);
    OpKmtGet:   StrCat(DlgCaption,TitKmtGet);
    OpKmtSend:  StrCat(DlgCaption,TitKmtSend);
    OpKmtFin:   StrCat(DlgCaption,TitKmtFin);
    OpXRcv:     StrCat(DlgCaption,TitXRcv);
    OpXSend:    StrCat(DlgCaption,TitXSend);
    OpZRcv:     StrCat(DlgCaption,TitZRcv);
    OpZSend:    StrCat(DlgCaption,TitZSend);
    OpBPRcv:     StrCat(DlgCaption,TitBPRcv);
    OpBPSend:    StrCat(DlgCaption,TitBPSend);
    OpQVRcv:     StrCat(DlgCaption,TitQVRcv);
    OpQVSend:    StrCat(DlgCaption,TitQVSend);
  end;
end;
end;

{Hook function for XMODEM file name dialog box}
function XFnHook(Dialog: HWnd; Message, WParam: Word; LParam: Longint): Bool; export;
var
  ofn: POpenFileName;
  Hi, Lo: word;
begin
  case Message of
    wm_InitDialog:
      begin
        ofn := POpenFileName(lParam);
        SetWindowLong(Dialog, DWL_USER, lParam);
        ShowDlgItem(Dialog,Id_FnBox,Id_FnX1K);
        SetRB(Dialog,HiWord(ofn^.lCustData),Id_FnXCheck,Id_FnX1K);
        if LoWord(ofn^.lCustData)<>$FFFF then
        begin
          ShowDlgItem(Dialog,Id_FnBinary,Id_FnBinary);
          SetRB(Dialog,LoWord(ofn^.lCustData),Id_FnBinary,Id_FnBinary);
        end;
        XFnHook := true;
        exit;
      end;
    wm_Command:
      case wParam of
        id_Ok: begin
          ofn := POpenFileName(GetWindowLong(Dialog,DWL_USER));
          if ofn<>nil then
          begin
            GetRB(Dialog,Hi,Id_FnXCheck,Id_FnX1K);
            if LoWord(ofn^.lCustData)=$FFFF then
              Lo := $FFFF
            else
              GetRB(Dialog,Lo,Id_FnBinary,Id_FnBinary);
            ofn^.lCustData := MakeLong(Lo,Hi);
          end;
        end;
        id_Cancel: ;
      end;
  end;
  XFnHook := False;
end;

function GetXFname(HWin: HWnd; Receive: bool; Option: Plongint; FileVar: PFileVar; CurDir: PChar): Bool; export;
var
  FNFilter: array [0..10] of Char;
  FNameRec: TOpenFileName;
  TempDir: array[0..fsDirectory] of char;
  Ok: BOOL;
begin
  GetXFname := false;
  {save current dir}
  GetCurDir(TempDir,0);

  FileVar^.FullName[0] := #0;
  FillChar(FNFilter, SizeOf(FNFilter), #0);  { Set up for double null at end }
  FillChar(FNameRec, SizeOf(TOpenFileName), #0);

  StrCopy(FileVar^.DlgCaption,'Tera Term: XMODEM ');
  if Receive then
    StrCat(FileVar^.DlgCaption,'Receive')
  else
    StrCat(FileVar^.DlgCaption,'Send');

  StrCopy(FNFilter, 'all');
  StrCopy(@FNFilter[StrLen(FNFilter)+1], '*.*');      

  with FNameRec do
  begin
    lStructSize   := sizeof(TOpenFileName);
    hwndOwner     := HWin;
    lpstrFilter   := FNFilter;
    nFilterIndex := 1;
    lpstrFile := FileVar^.FullName;
    nMaxFile := SizeOf(FileVar^.FullName);
    lpstrInitialDir := CurDir;
    flags := ofn_ShowHelp or ofn_HideReadOnly or
             ofn_EnableTemplate or ofn_EnableHook;
    lCustData := Option^;
    if not Receive then
    begin
      flags := flags or ofn_FileMustExist;
      lCustData := lCustData or $FFFF;
    end;
    lpstrTitle := FileVar^.DlgCaption;
    @lpfnHook := MakeProcInstance(@XFnHook, HInstance);
    lpTemplateName := PChar(Id_FnDlg);
  end;
  FNameRec.hInstance := hinstance;

  Ok := GetOpenFileName(FNameRec);

  if Ok then
  begin
    FileVar^.DirLen := FNameRec.nFileOffset;
    FileVar^.FnPtr := FNameRec.nFileOffSet;
    Move(FileVar^.FullName[0],CurDir[0],FileVar^.DirLen-1);
    CurDir[FileVar^.DirLen-1] := #0;

    if Receive then
      Option^ := FnameRec.lCustData
    else
      Option^ := MakeLong(LoWord(Option^),HiWord(FnameRec.lCustData));
  end;

  {restore dir}
  SetCurDir(TempDir);

  GetXFname := Ok;
end;

procedure ProtoInit
  (Proto: integer; fv: PFileVar; pv: pointer; cv: PComVar; ts: PTTSet); export;
begin
  case Proto of
    PROTO_KMT:
      KmtInit(fv,pv,cv,ts);
    PROTO_XM:
      XInit(fv,pv,cv,ts);
    PROTO_ZM:
      ZInit(fv,pv,cv,ts);
    PROTO_BP:
      BPInit(fv,pv,cv,ts);
    PROTO_QV:
      QVInit(fv,pv,cv,ts);
  end;
end;

function ProtoParse
  (Proto: integer; fv: PFileVar; pv: pointer; cv: PComVar): BOOL; export;
var
  Ok: BOOL;
begin
  case Proto of
    PROTO_KMT:
      Ok := KmtReadPacket(fv,pv,cv);
    PROTO_XM: begin
        case PXVar(pv)^.XMode of
	  IdXReceive:
	    Ok := XReadPacket(fv,pv,cv);
	  IdXSend:
	    Ok := XSendPacket(fv,pv,cv);
        end;
      end;
    PROTO_ZM:
      Ok := ZParse(fv,pv,cv);
    PROTO_BP:
      Ok := BPParse(fv,pv,cv);
    PROTO_QV: begin
        case PQVVar(pv)^.QVMode of
	  IdQVReceive:
	    Ok := QVReadPacket(fv,pv,cv);
	  IdQVSend:
	    Ok := QVSendPacket(fv,pv,cv);
        end;
      end;
  end;
  ProtoParse := Ok;
end;

procedure ProtoTimeOutProc(Proto: integer; fv: PFileVar; pv: pointer; cv: PComVar); export;
begin
  case Proto of
    PROTO_KMT:
      KmtTimeOutProc(fv,pv,cv);
    PROTO_XM:
      XTimeOutProc(fv,pv,cv);
    PROTO_ZM:
      ZTimeOutProc(fv,pv,cv);
    PROTO_BP:
      BPTimeOutProc(fv,pv,cv);
    PROTO_QV:
      QVTimeOutProc(fv,pv,cv);
  end;
end;

function ProtoCancel
  (Proto: integer; fv: PFileVar; pv: pointer; cv: PComVar): BOOL; export;
begin
  case Proto of
    PROTO_KMT:
      KmtCancel(fv,pv,cv);
    PROTO_XM:
      if PXVar(pv)^.XMode=IdXReceive then
        XCancel(fv,pv,cv);
    PROTO_ZM:
      ZCancel(pv);
    PROTO_BP:
      if PBPVar(pv)^.BPState <> BP_Failure then
      begin
	BPCancel(pv);
        ProtoCancel := FALSE;
        exit;
      end;
    PROTO_QV:
      QVCancel(fv,pv,cv);
  end;
  ProtoCancel := TRUE;
end;

exports
  GetSetupFname      index 1, 
  GetTransFname      index 2,
  GetMultiFname      index 3,
  GetGetFname        index 4,
  SetFileVar         index 5,
  GetXFname          index 6,

  ProtoInit          index 7,
  ProtoParse         index 8,
  ProtoTimeOutProc   index 9,
  ProtoCancel        index 10,

  VTPrint            index 20;

begin
end.
