{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TTMACRO.EXE, Tera Term Language interpreter}
unit TTL;

interface
{$I teraterm.inc}

{$IFDEF Delphi}
uses Messages, WinTypes, WinProcs, OWindows, Strings, WinDos, TTMDLG,
     TTMBuff, TTMParse, TTMDDE;
{$ELSE}
uses WinTypes, WinProcs, WObjects, Strings, WinDos, TTMDLG,
     TTMBuff, TTMParse, TTMDDE;
{$ENDIF}

{$DEFINE TTERM16} { <-- TTERM16 or TTERM32 }
const
{$IFDEF TTERM32}
  TTERMCOMMAND = 'TTERMPRO /D=';
{$ELSE}
  TTERMCOMMAND = 'TERATERM /D=';
{$ENDIF}

const
  IdTimeOutTimer = 1;

var
  TimeLimit: longint;

function InitTTL: boolean;
procedure EndTTL;
function ExecCmnd: integer;
procedure Exec;
procedure SetInputStr(Str: PChar);
procedure SetResult(ResultCode: integer);
procedure SetTimeOut(TimeOut: integer);
function CalcTime: longint;

implementation

{for 'ExecCmnd' command}
var
  ParseAgain: BOOL;
  IfNest: integer;
  ElseFlag: integer;
  EndIfFlag: integer;

function InitTTL: boolean;
var
  i, j: integer;
  Temp: TStrVal;
begin
  InitTTL := false;
  InitVar;
  LockVar;

  {System variables}
  NewIntVar('result',0);
  NewIntVar('timeout',0);
  NewStrVar('inputstr',#0);

  j := 0;
  for i := 1 to ParamCount do
  begin
    StrPCopy(Temp,ParamStr(i));
    if strlicomp(Temp,'/D=',3)<>0 then
    begin
      inc(j);
      if j=2 then
        NewStrVar('param2',Temp)
      else if j=3 then
        NewStrVar('param3',Temp);
    end;
  end;
  Temp[0] := #0;
  if j<2 then
    NewStrVar('param2',Temp);
  if j<3 then
    NewStrVar('param3',Temp);

  ParseAgain := FALSE;
  IfNest := 0;
  ElseFlag := 0;
  EndIfFlag := 0;

  if not InitBuff(FileName) then
  begin
    TTLStatus := IdTTLEnd;
    exit;
  end;

  UnlockVar;
  TTLStatus := IdTTLRun;
  InitTTL := TRUE;
end;

procedure EndTTL;
var
  i: integer;
begin
  UnlockVar;
  if TTLStatus=IdTTLWait then
    KillTimer(Application^.MainWindow^.HWindow,IdTimeOutTimer);
  CloseBuff(0);
  EndVar;
end;

{from here, TTL commands}
function TTLBeep: word;
begin
  if GetFirstChar=#0 then
  begin
    TTLBeep := 0;
    messagebeep(0);
  end
  else
    TTLBeep := ErrSyntax;
end;

function TTLBPlusRecv: word;
begin
  if GetFirstChar<>#0 then
    TTLBPlusRecv := ErrSyntax
  else
    TTLBPlusRecv := SendCmnd(CmdBPlusRecv,true);
end;

function TTLBPlusSend: word;
var
  Str: TStrVal;
  Err: word;
begin
  Err := 0;
  GetStrVal(Str,Err);

  if (Err=0) and
     ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;

  TTLBPlusSend := Err;
  if Err=0 then
  begin
    SetFile(Str);
    TTLBPlusSend := SendCmnd(CmdBPlusSend,true);
  end;
end;

function TTLCall: word;
var
  LabName: TName;
  Err, VarType, VarId: word;
begin
  if GetLabelName(LabName) and (GetFirstChar=#0) then
  begin
    if CheckVar(LabName,VarType,VarId) and (VarType=TypLabel) then
      Err := CallToLabel(VarId)
    else
      Err := ErrLabelReq;
  end
  else
    Err := ErrSyntax;

  TTLCall := Err;
end;

function TTLChangeDir: word;
var
  Str: TStrVal;
  Err: word;
begin
  Err := 0;
  GetStrVal(Str,Err);

  if (Err=0) and
     ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;

  TTLChangeDir := Err;
  if Err=0 then
  begin
    StrUpper(Str);
    SetFile(Str);
    TTLChangeDir := SendCmnd(CmdChangeDir,false);
  end;
end;

function TTLCloseTT: word;
begin
  if GetFirstChar<>#0 then
  begin
    TTLCloseTT := ErrSyntax;
    exit;
  end;
  if ConvH = 0 then   
    TTLCloseTT := ErrLinkFirst
  else begin
  {Close Tera Term}
    TTLCloseTT := SendCmnd(CmdCloseWin,FALSE);
    TTLStatus := IdTTLEnd;
    TTLCloseTT := 0;
  end;
end;

function TTLConnect: word;
var
  Cmnd, Str: TStrVal;
  Err: word;
  Msg: TMsg;
  w: word;
begin
  Str[0] := #0;

  Err := 0;
  GetStrVal(Str,Err);

  TTLConnect := Err;
  if Err<>0 then exit;

  if GetFirstChar<>#0 then
  begin
    TTLConnect := ErrSyntax;
    exit;
  end;

  if ConvH<>0 then exit;

  if strlen(TopicName)=0 then
  begin
    strcopy(Cmnd,TTERMCOMMAND);
    w := Application^.MainWindow^.HWindow;
    Byte2HexStr(Hi(w),TopicName);
    Byte2HexStr(Lo(w),@TopicName[2]);
    strcat(Cmnd,TopicName);
    strcat(Cmnd,' ');
    StrLCat(Cmnd,Str,SizeOf(Cmnd)-1);
    TTLStatus := IdTTLInitDDE;
    if WinExec(Cmnd,sw_show)<32 then
      TTLConnect := ErrCantConnect;
  end;

end;

function TTLElse: word;
begin
  if GetFirstChar<>#0 then
  begin
    TTLElse := ErrSyntax;
    exit;
  end;
  if IfNest<1 then
  begin
    TTLElse := ErrInvalidCtl;
    exit;
  end;  
  {Skip until 'EndIf}
  dec(IfNest);
  EndIfFlag := 1;
  TTLElse := 0;
end;

function CheckElseIf(var Err: word): integer;
var
  Val: integer;
  WId: word;
begin
  CheckElseIf := 0;
  Err := 0;
  GetIntVal(Val,Err);
  if Err<>0 then exit;
  CheckElseIf := Val;
  if not GetReservedWord(WId) or
     (WId<>RsvThen) or
     (GetFirstChar<>#0) then
    Err := ErrSyntax;
end;

function TTLElseIf: word;
var
  Err: word;
  Val: integer;
begin
  Val := CheckElseIf(Err);
  TTLElseIf := Err;
  if Err<>0 then exit;

  if IfNest<1 then
  begin
    TTLElseIf := ErrInvalidCtl;
    exit;
  end;
  {Skip until 'EndIf'}
  dec(IfNest);
  EndIfFlag := 1;
end;

function TTLEnd: word;
begin
  if GetFirstChar=#0 then
  begin
    TTLEnd := 0;
    TTLStatus := IdTTLEnd;
  end
  else
    TTLEnd := ErrSyntax;
end;

function TTLEndIf: word;
begin
  if GetFirstChar<>#0 then
  begin
    TTLEndIf := ErrSyntax;
    exit;
  end;
  if IfNest<1 then
  begin
    TTLEndIf := ErrInvalidCtl;
    exit;
  end;  
  dec(IfNest);
  TTLEndIf := 0;
end;

function TTLEndWhile: word;
begin
  if GetFirstChar<>#0 then
  begin
    TTLEndWhile := ErrSyntax;
    exit;
  end;
  TTLEndWhile := BackToWhile;
end;

function TTLExec: word;
var
  Str: TStrVal;
  Err: word;
begin
  Err := 0;
  GetStrVal(Str,Err);

  if (Err=0) and
     ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;

  TTLExec := Err;
  if Err<>0 then exit;

  WinExec(Str,sw_show);
end;

function TTLExecCmnd: word;
var
  Err: word;
  NextLine: TStrVal;
  b: char;
begin
  Err := 0;
  GetStrVal(NextLine,Err);
  TTLExecCmnd := Err;
  if Err<>0 then exit;
  if GetFirstChar<>#0 then
  begin
    TTLExecCmnd := ErrSyntax;
    exit;
  end;

  strcopy(LineBuff,NextLine);
  LineLen := strlen(LineBuff);
  LinePtr := 0;
  b := GetFirstChar;
  dec(LinePtr);
  ParseAgain := (b<>#0) and (b<>':') and (b<>';')
end;

function TTLExit: word;
begin
  if GetFirstChar=#0 then
  begin
    TTLExit := 0;
    if not ExitBuffer then
      TTLStatus := IdTTLEnd;
  end
  else
    TTLExit := ErrSyntax;
end;

function TTLFileClose: word;
var
  Err: word;
  FH: integer;
begin
  Err := 0;
  GetIntVal(FH,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLFileClose := Err;
  if Err<>0 then exit;
  _lclose(FH);
end;

function TTLFileConcat: word;
var
  Err: word;
  FH1, FH2, c: integer;
  FName1, FName2: TStrVal;
  buf: array[0..1023] of byte;
begin
  Err := 0;
  GetStrVal(FName1,Err);
  GetStrVal(FName2,Err);
  if (Err=0) and
    ((StrLen(FName1)=0) or
     (StrLen(FName2)=0) or
     (GetFirstChar<>#0)) then
    Err := ErrSyntax;
  TTLFileConcat := Err;
  if Err<>0 then exit;
  if StrIComp(FName1,FName2)=0 then exit;

  FH1 := _lopen(FName1,of_Write);
  if FH1<0 then
    FH1 := _lcreat(FName1,0);
  if FH1<0 then exit;
  _llseek(FH1,0,2);

  FH2 := _lopen(FName2,of_Read);
  if FH2<>-1 then
  begin
    repeat
      c := _lread(FH2,@buf[0],SizeOf(buf));
      if c>0 then
        _lwrite(FH1,@buf[0],c);
    until c < SizeOf(buf);
    _lclose(FH2);
  end;
  _lclose(FH1);
end;

function TTLFileCopy: word;
var
  Err: word;
  FH1, FH2, c: integer;
  FName1, FName2: TStrVal;
  buf: array[0..1023] of byte;
begin
  Err := 0;
  GetStrVal(FName1,Err);
  GetStrVal(FName2,Err);
  if (Err=0) and
    ((StrLen(FName1)=0) or
     (StrLen(FName2)=0) or
     (GetFirstChar<>#0)) then
    Err := ErrSyntax;
  TTLFileCopy := Err;
  if Err<>0 then exit;
  if StrIComp(FName1,FName2)=0 then exit;

  FH1 := _lopen(FName1,of_Read);
  if FH1<0 then exit;
  FH2 := _lcreat(FName2,0);
  if FH2<>-1 then
  begin
    repeat
      c := _lread(FH1,@buf[0],SizeOf(buf));
      if c>0 then
        _lwrite(FH2,@buf[0],c);
    until c < SizeOf(buf);
    _lclose(FH2);
  end;
  _lclose(FH1);
end;

function TTLFileCreate: word;
var
  Err, VarId: word;
  FH: integer;
  FName: TStrVal;
begin
  Err := 0;
  GetIntVar(VarId,Err);
  GetStrVal(FName,Err);
  if (Err=0) and
    ((StrLen(FName)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;
  TTLFileCreate := Err;
  if Err<>0 then exit;

  FH := _lcreat(FName,0);
  if FH<0 then FH := -1;
  IntVal[VarId] := FH;
end;

function TTLFileDelete: word;
var
  Err: word;
  FName: TStrVal;
  f: file;
begin
  Err := 0;
  GetStrVal(FName,Err);
  if (Err=0) and
    ((StrLen(FName)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;
  TTLFileDelete := Err;
  if Err<>0 then exit;

  Assign(f,StrPas(FName));
  {$I-}
  Reset(f);
  {$I+}
  if IOResult=0 then
  begin
    Close(f);
    Erase(f);
  end;
end;

function TTLFileOpen: word;
var
  Err, VarId: word;
  Append, FH: integer;
  FName: TStrVal;
begin
  Err := 0;
  GetIntVar(VarId,Err);
  GetStrVal(FName,Err);
  GetIntVal(Append,Err);
  if (Err=0) and
    ((StrLen(FName)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;
  TTLFileOpen := Err;
  if Err<>0 then exit;

  FH := _lopen(FName,of_ReadWrite);
  if FH<0 then
    FH := _lcreat(FName,0);
  if FH<0 then FH := -1;
  IntVal[VarId] := FH;
  if FH<0 then exit;
  if Append<>0 then _llseek(FH,0,2);  
end;

function TTLFileReadln: word;
var
  Err, VarId: word;
  FH, i, c: integer;
  Str: TStrVal;
  EOF, EOL: boolean;
  b: byte;
begin
  Err := 0;
  GetIntVal(FH,Err);
  GetStrVar(VarId,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLFileReadln := Err;
  if Err<>0 then exit;

  i := 0;
  EOL := false;
  EOF := true;
  repeat
    c := _lread(FH,@b,1);
    if c>0 then EOF := false;
    if c=1 then
      case b of
        $0d: begin
            c := _lread(FH,@b,1);
            if (c=1) and (b<>$0a) then
              _llseek(FH,-1,1);
            EOL := true;
          end;
        $0a: EOL := true;
      else
        if i<MaxStrLen then
        begin          
          Str[i] := char(b);
          inc(i);
        end;
      end;
  until (c<1) or EOL;

  if EOF then
    SetResult(1)
  else
    SetResult(0);

  Str[i] := #0;
  SetStrVal(VarId,Str);
end;

function TTLFileRename: word;
var
  Err: word;
  FName1, FName2: TStrVal;
  f: file;
begin
  Err := 0;
  GetStrVal(FName1,Err);
  GetStrVal(FName2,Err);
  if (Err=0) and
    ((StrLen(FName1)=0) or
     (StrLen(FName2)=0) or
     (GetFirstChar<>#0)) then
    Err := ErrSyntax;
  TTLFileRename := Err;
  if Err<>0 then exit;
  if StrIComp(FName1,FName2)=0 then exit;
  {$I-}
  Assign(f,StrPas(FName1));
  Rename(f,StrPas(FName2));
  {$I+}
end;

function TTLFileSearch: word;
var
  Err: word;
  FName: TStrVal;
  SearchRec: TSearchRec;
begin
  Err := 0;
  GetStrVal(FName,Err);
  if (Err=0) and
    ((StrLen(FName)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;
  TTLFileSearch := Err;
  if Err<>0 then exit;

  FindFirst(FName,faAnyFile,SearchRec);
  if DosError=0 then
    SetResult(1)
  else
    SetResult(0);
end;
  
function TTLFileSeek: word;
var
  Err: word;
  FH, i, j: integer;
begin
  Err := 0;
  GetIntVal(FH,Err);
  GetIntVal(i,Err);
  GetIntVal(j,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLFileSeek := Err;
  if Err<>0 then exit;
  _llseek(FH,i,j);
end;

function TTLFileStrSeek: word;
var
  Err: word;
  FH, Len, i, c: integer;
  Str: TStrVal;
  b: byte;
  pos: longint;
begin
  Err := 0;
  GetIntVal(FH,Err);
  GetStrVal(Str,Err);
  if (Err=0) and
    ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;
  TTLFileStrSeek := Err;
  if Err<>0 then exit;
  pos := _llseek(FH,0,1);
  if pos=-1 then exit;

  Len := StrLen(Str);
  i := 0;
  repeat
    c := _lread(FH,@b,1);
    if c=1 then
    begin
      if b=byte(Str[i]) then
        inc(i)
      else
        i := 0;
    end;
  until (c<1) or (i=Len);
  if i=Len then
    SetResult(1)
  else begin
    SetResult(0);
    _llseek(FH,pos,0);
  end;
end;

function TTLFileWrite: word;
var
  Err: word;
  FH: integer;
  Str: TStrVal;
begin
  Err := 0;
  GetIntVal(FH,Err);
  GetStrVal(Str,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLFileWrite := Err;
  if Err<>0 then exit;

  _lwrite(FH,Str,StrLen(Str));
end;

function TTLFileWriteln: word;
var
  Err: word;
  FH: integer;
  Str: TStrVal;
begin
  Err := 0;
  GetIntVal(FH,Err);
  GetStrVal(Str,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLFileWriteln := Err;
  if Err<>0 then exit;

  _lwrite(FH,Str,StrLen(Str));
  _lwrite(FH,#$0d#$0a,2);
end;

function TTLFor: word;
var
  Err, VarId: word;
  ValStart, ValEnd: integer;
begin
  Err := 0;
  GetIntVar(VarId,Err);
  GetIntVal(ValStart,Err);
  GetIntVal(ValEnd,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLFor := Err;
  if Err<>0 then exit;

  if not CheckNext then
  begin {first time}
    TTLFor := SetForLoop;
    IntVal[VarId] := ValStart;
  end
  else begin {return from 'next'}
    if IntVal[VarId]<ValEnd then
      inc(IntVal[VarId])
    else if IntVal[VarId]>ValEnd then
      dec(IntVal[VarId]);
    if IntVal[VarId]=ValEnd then
      LastForLoop;
  end;
end;

procedure AddDate(DateStr: PChar; DateNum: word);
var
  NumStr: string[10];
  PasStr: array[0..10] of char;
begin
  Str(DateNum,NumStr);
  StrPCopy(PasStr,NumStr);
  if StrLen(PasStr)=1 then
    StrCat(DateStr,'0');
  StrCat(DateStr,PasStr);
end;

function TTLGetDate: word;
var
  VarId, Err: word;
  Str2: TStrVal;
  y, m, d, dw: word;
begin
  Err := 0;
  GetStrVar(VarId,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLGetDate := Err;
  if Err<>0 then exit;

  GetDate(y,m,d,dw);
  Str2[0] := #0;
  AddDate(Str2,y);
  StrCat(Str2,'-');
  AddDate(Str2,m);
  StrCat(Str2,'-');
  AddDate(Str2,d);

  SetStrVal(VarId,Str2);
end;

function TTLGetTime: word;
var
  VarId, Err: word;
  Str2: TStrVal;
  h, m, s, hund: word;
begin
  Err := 0;
  GetStrVar(VarId,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLGetTime := Err;
  if Err<>0 then exit;

  GetTime(h,m,s,hund);
  Str2[0] := #0;
  AddDate(Str2,h);
  StrCat(Str2,':');
  AddDate(Str2,m);
  StrCat(Str2,':');
  AddDate(Str2,s);

  SetStrVal(VarId,Str2);
end;

function TTLGoto: word;
var
  LabName: TName;
  Err, VarType, VarId: word;
begin
  if GetLabelName(LabName) and (GetFirstChar=#0) then
  begin
    if CheckVar(LabName,VarType,VarId) and (VarType=TypLabel) then
    begin
      JumpToLabel(VarId);
      Err := 0;
    end
    else
      Err := ErrLabelReq;
  end
  else
    Err := ErrSyntax;

  TTLGoto := Err;
end;

function CheckThen(var Err: word): BOOL;
var
  b: char;
  Temp: TName;
begin
  CheckThen := false;

  repeat

    repeat
      b := GetFirstChar;
      if b=#0 then exit;
    until ((b>='A') and (b<='Z')) or
          (b='_') or
          ((b>='a') and (b<='z'));
    dec(LinePtr);
    if not GetIdentifier(Temp) then exit;

  until stricomp(Temp,'then')=0;

  CheckThen := true;
  if GetFirstChar<>#0 then
    Err := ErrSyntax;
end;

function TTLIf: word;
var
  Err, ValType, Tmp, WId: word;
  Val: integer;
begin
  if not GetExpression(ValType,Val,Err) then
  begin
    TTLIf := ErrSyntax;
    exit;
  end;

  TTLIf := Err;
  if Err<>0 then exit;

  if ValType<>TypInteger then
  begin
    TTLIf := ErrTypeMismatch;
    exit;
  end;

  Tmp := LinePtr;
  if GetReservedWord(WId) and
     (WId=RsvThen) then
  begin  {If then ... EndIf structure}
    if GetFirstChar<>#0 then
    begin
      TTLIf := ErrSyntax;
      exit;
    end;
    inc(IfNest); 
    if Val=0 then
      ElseFlag := 1; {Skip until 'Else' or 'EndIf'}
  end
  else begin {single line lf command}
    if Val=0 then
    begin
      TTLIf := 0;
      exit;
    end;
    LinePtr := Tmp;
    TTLIf := ExecCmnd;
    exit;
  end;
end;

function TTLInclude: word;
var
  Err: word;
  Str: TStrVal;
begin
  Err := 0;
  GetStrVal(Str,Err);
  if not BuffInclude(Str) then
    Err := ErrCantOpen;
  TTLInclude := Err;
end;

function TTLInputBox(Paswd: BOOL): word;
var
  Str1, Str2, InStr: TStrVal;
  Err, ValType, VarId: word;
  Val, i: integer;
begin
  Err := 0;
  GetStrVal(Str1,Err);
  GetStrVal(Str2,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;

  TTLInputBox := Err;
  if Err<>0 then exit;

  if CheckVar('inputstr',ValType,VarId) and
     (ValType=TypString) then
    Application^.ExecDialog(New(PInpDlg,
        Init(StrVarPtr(VarId),Str1,Str2,Paswd) ));

end;

function TTLInt2Str: word;
var
  VarId, Err: word;
  Num: integer;
  Str2: TStrVal;
  NumStr: String[10];
begin
  Err := 0;
  GetStrVar(VarId,Err);

  GetIntVal(Num,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLInt2Str := Err;
  if Err<>0 then exit;

  Str(Num,NumStr);
  StrPCopy(Str2,NumStr);

  SetStrVal(VarId,Str2);
end;

function TTLKmtRecv: word;
begin
  if GetFirstChar<>#0 then
    TTLKmtRecv := ErrSyntax
  else
    TTLKmtRecv := SendCmnd(CmdKmtRecv,true);
end;

function TTLKmtSend: word;
var
  Str: TStrVal;
  Err: word;
begin
  Err := 0;
  GetStrVal(Str,Err);

  if (Err=0) and
     ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;

  TTLKmtSend := Err;
  if Err=0 then
  begin
    SetFile(Str);
    TTLKmtSend := SendCmnd(CmdKmtSend,true);
  end;
end;

function TTLLogClose: word;
begin
  if GetFirstChar<>#0 then
    TTLLogClose := ErrSyntax
  else
    TTLLogClose := SendCmnd(CmdLogClose,false);
end;

function TTLLogOpen: word;
var
  Str: TStrVal;
  Err: word;
  BinFlag, AppendFlag: integer;
begin
  Err := 0;
  GetStrVal(Str,Err);
  GetIntVal(BinFlag,Err);
  GetIntVal(AppendFlag,Err);
  if (Err=0) and
    ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;

  TTLLogOpen := Err;
  if Err<>0 then exit;

  SetFile(Str);
  SetBinary(BinFlag);
  SetAppend(AppendFlag);
  TTLLogOpen := SendCmnd(CmdLogOpen,false);
end;

function TTLLogPause: word;
begin
  if GetFirstChar<>#0 then
    TTLLogPause := ErrSyntax
  else
    TTLLogPause := SendCmnd(CmdLogPause,false);
end;

function TTLLogStart: word;
begin
  if GetFirstChar<>#0 then
    TTLLogStart := ErrSyntax
  else
    TTLLogStart := SendCmnd(CmdLogStart,false);
end;

function TTLLogWrite: word;
var
  Str: TSTrVal;
  Err: word;
begin
  Err := 0;
  GetStrVal(Str,Err);

  if (Err=0) and
     ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;

  TTLLogWrite := Err;
  if Err<>0 then exit;

  SetFile(Str);
  TTLLogWrite := SendCmnd(CmdLogWrite,false);
end;

function MessageCommand(MB: word; var Err: word): integer;
var
  Str1, Str2: TStrVal;
begin
  Err := 0;
  GetStrVal(Str1,Err);
  GetStrVal(Str2,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  if Err<>0 then exit;

  MessageCommand :=
    Messagebox(Application^.MainWindow^.HWindow,
               Str1, Str2, MB);  
end;

function TTLMessageBox: word;
var
  Err: word;
begin
  MessageCommand(MB_OK, Err);
  TTLMessageBox := Err;
end;

function TTLNext: word;
begin
  if GetFirstChar<>#0 then
  begin
    TTLNext := ErrSyntax;
    exit;
  end;
  TTLNext := NextLoop;
end;

function TTLPause: word;
var
  TimeOut: integer;
  Err: word;
begin
  Err := 0;
  GetIntVal(TimeOut,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLPause := Err;
  if Err<>0 then exit;

  if TimeOut>0 then
  begin
    TTLStatus := IdTTLPause;
    TimeLimit := CalcTime + TimeOut;
    if TimeLimit>=86400 then TimeLimit := TimeLimit-86400;
    SetTimer(Application^.MainWindow^.HWindow,
             IdTimeOutTimer,1000, nil)
  end

end;

function TTLQuickVANRecv: word;
begin
  if GetFirstChar<>#0 then
    TTLQuickVANRecv := ErrSyntax
  else
    TTLQuickVANRecv := SendCmnd(CmdQVRecv,true);
end;

function TTLQuickVANSend: word;
var
  Str: TStrVal;
  Err: word;
begin
  Err := 0;
  GetStrVal(Str,Err);

  if (Err=0) and
     ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;

  TTLQuickVANSend := Err;
  if Err=0 then
  begin
    SetFile(Str);
    TTLQuickVANSend := SendCmnd(CmdQVSend,true);
  end;
end;

function TTLReturn: word;
begin
  if GetFirstChar=#0 then
    TTLReturn := ReturnFromSub
  else
    TTLReturn := ErrSyntax;
end;

function TTLSend: word;
var
  Str: TStrVal;
  Err, ValType: word;
  Val: integer;
  EndOfLine: boolean;
begin
  if ConvH=0 then
  begin
    TTLSend := ErrLinkFirst;
    exit;
  end;

  EndOfLine := false;
  repeat
    if GetString(Str,Err) then
    begin
      TTLSend := Err;
      if Err<>0 then exit;
      DDEOut(Str);
    end
    else if GetExpression(ValType,Val,Err) then
    begin
      TTLSend := Err;
      if Err<>0 then exit;
      case ValType of
        TypInteger: DDEOut1Byte(byte(Val));
        TypString: DDEOut(StrVarPtr(Val));
      else
        begin
          TTLSend := ErrTypeMismatch;
          exit;
        end;
      end;
    end
    else
      EndOfLine := true;
    TTLSend := 0;
  until EndOfLine;
end;

function TTLSendFile: word;
var
  Str: TStrVal;
  Err: word;
  BinFlag: integer;
begin
  Err := 0;
  GetStrVal(Str,Err);
  GetIntVal(BinFlag,Err);
  if (Err=0) and
    ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;

  TTLSendFile := Err;
  if Err<>0 then exit;

  SetFile(Str);
  SetBinary(BinFlag);
  TTLSendFile := SendCmnd(CmdSendFile,true);
end;

function TTLSendln: word;
var
  Err: word;
  Str: array[0..2] of char;
begin
  Err := TTLSend;
  if Err=0 then
  begin
    Str[0] := #$0D;
    Str[1] := #$0A;
    Str[2] := #0;
    if ConvH<>0 then
      DDEOut(Str)
    else
      Err := ErrLinkFirst;
  end;
  TTLSendln := Err;
end;

function TTLShow(TT: BOOL): word;
var
  Err: word;
  Val: integer;
begin
  Err := 0;
  GetIntVal(Val,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLShow := Err;
  if Err<>0 then exit;
  if TT then
  begin
    if (ConvH=0) then
    begin
      TTLShow := ErrLinkFirst;
      exit;
    end;
    if Val=0 then
      TTLShow := SendCmnd(CmdShowMin,FALSE)
    else
      TTLShow := SendCmnd(CmdShowRestore,FALSE);
  end
  else begin 
    if Val=0 then
      ShowWindow(HWin,SW_MINIMIZE)
    else
      ShowWindow(HWin,SW_RESTORE);
  end;
end;

function TTLStr2Int: word;
var
  VarId, Err: word;
  Str: TStrVal;
  Num, c: integer;
begin
  Err := 0;
  GetIntVar(VarId,Err);
  GetStrVal(Str,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;

  TTLStr2Int := Err;
  if Err<>0 then exit;

  Val(StrPas(Str),Num,c);
  if c<>0 then
  begin
    Num := 0;
    SetResult(0);
  end
  else
    SetResult(1);
  IntVal[VarId] := Num;
end;

function TTLStrCompare: word;
var
  Str1, Str2: TStrVal;
  Err: word;
  i: integer;
begin
  Err := 0;
  GetStrVal(Str1,Err);
  GetStrVal(Str2,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLStrCompare := Err;
  if Err<>0 then exit;

  i := StrComp(Str1,Str2);
  if i<0 then
    i := -1
  else if i>0 then
    i := 1;
  SetResult(i);
end;

function TTLStrConcat: word;
var
  VarId, Err: word;
  Str: TStrVal;
begin
  Err := 0;
  GetStrVar(VarId,Err);
  GetStrVal(Str,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLStrConcat := Err;
  if Err<>0 then exit;

  StrLCat(StrVarPtr(VarId),Str,MaxStrLen);
end;

function TTLStrCopy: word;
var
  Err, VarId: word;
  From, Len, SrcLen: integer;
  Str: TStrVal;
begin
  Err := 0;
  GetStrVal(Str,Err);
  GetIntVal(From,Err);
  GetIntVal(Len,Err);
  GetStrVar(VarId,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLStrCopy := Err;
  if Err<>0 then exit;

  if From<1 then From := 1;
  SrcLen := StrLen(Str)-From+1;
  if Len > SrcLen then Len := SrcLen;
  if Len < 0 then Len := 0;
  Move(Str[From-1],StrVarPtr(VarId)[0],Len);
  StrVarPtr(VarId)[Len] := #0;
end;

function TTLStrLen: word;
var
  Err: word;
  Str: TStrVal;
begin
  Err := 0;
  GetStrVal(Str,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLStrLen := Err;
  if Err<>0 then exit;
  SetResult(StrLen(Str));
end;

function TTLStrScan: word;
var
  Err: word;
  Len1, Len2, i, j: integer;
  Str1, Str2: TStrVal;
begin
  Err := 0;
  GetStrVal(Str1,Err);
  GetStrVal(Str2,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLStrScan := Err;
  if Err<>0 then exit;

  Len1 := StrLen(Str1);
  Len2 := StrLen(Str2);
  if (Len1=0) or (Len2=0) then
  begin
    SetResult(0);
    exit;
  end;

  i := 0;
  j := 0;
  repeat
    if Str1[i]=Str2[j] then
      inc(j)
    else
      j := 0;
    inc(i);
  until (i=Len1) or (j=Len2);
  if j=Len2 then
    SetResult(i-Len2+1)
  else
    SetResult(0);
end;

function TTLWait: word;
var
  Str: TStrVal;
  Err, ValType: word;
  i, Val: integer;
  NoMore: boolean;
  TimeOut: integer;
begin
  ClearWait;

  i := 0;
  repeat
    Err := 0;
    Str[0] := #0;
    NoMore := false;
    if not GetString(Str,Err) then
    begin
      if GetExpression(ValType,Val,Err) then
      begin
        if Err=0 then
        begin
          if ValType=TypString then
            StrCopy(Str,StrVarPtr(Val))
          else
            Err:=ErrTypeMismatch;
        end;
      end
      else
        NoMore := true;
    end;

    if (Err=0) and (StrLen(Str)>0) and (i<10) then
    begin
      inc(i);
      SetWait(i,Str);
    end;
  until (Err<>0) or (i>=10) or NoMore;

  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;

  if ConvH=0 then
    Err := ErrLinkFirst;

  if (Err=0) and (i>0) then
  begin
    TTLStatus := IdTTLWait;
    TimeOut := 0;
    if CheckVar('timeout',ValType,word(Val)) and
     (ValType=TypInteger) then
      TimeOut := IntVal[Val];

    if TimeOut>0 then
    begin
      TimeLimit := CalcTime + TimeOut;
      if TimeLimit>=86400 then TimeLimit := TimeLimit-86400;
      SetTimer(Application^.MainWindow^.HWindow,
               IdTimeOutTimer,1000, nil)
    end;
  end
  else
    ClearWait;

  TTLWait := Err;
end;

function TTLWaitRecv: word;
var
  Str: TStrVal;
  Err: word;
  Pos, Len, TimeOut: integer;
  VarType, VarId: word;
begin
  Err := 0;
  GetStrVal(Str,Err);
  GetIntVal(Len,Err);
  GetIntVal(Pos,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  if (Err=0) and (ConvH=0) then
    Err := ErrLinkFirst;
  TTLWaitRecv := Err;
  if Err<>0 then exit;

  SetWait2(Str,Len,Pos);

  TTLStatus := IdTTLWait2;
  TimeOut := 0;
  if CheckVar('timeout',VarType,VarId) and
     (VarType=TypInteger) then
    TimeOut := IntVal[VarId];
  if TimeOut>0 then
  begin
    TimeLimit := CalcTime + TimeOut;
    if TimeLimit>=86400 then TimeLimit := TimeLimit-86400;
    SetTimer(Application^.MainWindow^.HWindow,
             IdTimeOutTimer,1000, nil)
  end;

end;

function TTLWhile: word;
var
  Err: word;
  Val: integer;
begin
  Err := 0;
  GetIntVal(Val,Err);
  if (Err=0) and (GetFirstChar<>#0) then
    Err := ErrSyntax;
  TTLWhile := Err;
  if (Err<>0) then exit;

  if Val<>0 then
    TTLWhile := SetWhileLoop
  else
    EndWhileLoop;
end;

function TTLXmodemRecv: word;
var
  Str: TStrVal;
  Err: word;
  BinFlag, XOption: integer;
begin
  Err := 0;
  GetStrVal(Str,Err);
  GetIntVal(BinFlag,Err);
  GetIntVal(XOption,Err);
  if (Err=0) and
    ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;

  TTLXmodemRecv := Err;
  if Err<>0 then exit;

  SetFile(Str);
  SetBinary(BinFlag);
  SetXOption(XOption);
  TTLXmodemRecv := SendCmnd(CmdXmodemRecv,true);
end;

function TTLXmodemSend: word;
var
  Str: TStrVal;
  Err: word;
  XOption: integer;
begin
  Err := 0;
  GetStrVal(Str,Err);
  GetIntVal(XOption,Err);
  if (Err=0) and
    ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;
  TTLXmodemSend := Err;
  if Err<>0 then exit;

  SetFile(Str);
  SetXOption(XOption);
  TTLXmodemSend := SendCmnd(CmdXmodemSend,true);
end;

function TTLYesNoBox: word;
var
  Err: word;
  YesNo: integer;
begin
  YesNo := MessageCommand(MB_YesNo, Err);
  TTLYesNoBox := Err;
  if Err<>0 then exit;
  if YesNo=IDYes then
    YesNo := 1   {Yes}
  else
    YesNo := 0;  {No} 
  SetResult(YesNo);
end;

function TTLZmodemRecv: word;
begin
  if GetFirstChar<>#0 then
    TTLZmodemRecv := ErrSyntax
  else
    TTLZmodemRecv := SendCmnd(CmdZmodemRecv,true);
end;

function TTLZmodemSend: word;
var
  Str: TStrVal;
  Err: word;
  BinFlag: integer;
begin
  Err := 0;
  GetStrVal(Str,Err);
  GetIntVal(BinFlag,Err);
  if (Err=0) and
    ((StrLen(Str)=0) or (GetFirstChar<>#0)) then
    Err := ErrSyntax;
  TTLZmodemSend := Err;
  if Err<>0 then exit;

  SetFile(Str);
  SetBinary(BinFlag);
  TTLZmodemSend := SendCmnd(CmdZmodemSend,true);
end;

function ExecCmnd: integer;
var
  WId, Err: word;
  StrConst, E: boolean;
  Str: TStrVal;
  Cmnd: TName;
  ValType, VarType, VarId: word;
  Val: integer;
begin
  if (EndWhileFlag>0) then
  begin
    if not GetReservedWord(WId) then
      WId := 0;
    if WId=RsvWhile then
      inc(EndWhileFlag)
    else if WId=RsvEndWhile then
      dec(EndWhileFlag);
    ExecCmnd := 0;
    exit;
  end;

  if (EndIfFlag>0) then
  begin
    Err := 0;
    if not GetReservedWord(WId) then
      WId := 0;
    if (WId=RsvIf) and CheckThen(Err) then
      inc(EndIfFlag)
    else if (WId=RsvEndIf) then
      dec(EndIfFlag);
    ExecCmnd := Err;
    exit;
  end;

  if (ElseFlag>0) then
  begin
    Err := 0;
    if not GetReservedWord(WId) then
      WId := 0;
    if (WId=RsvIf) and CheckThen(Err) then      
      inc(EndIfFlag)
    else if (WId=RsvElse) then
      dec(ElseFlag)
    else if (WId=RsvElseIf) then
    begin
      if CheckElseIf(Err)<>0 then
        dec(ElseFlag);
    end
    else if (WId=RsvEndIf) then
    begin
      dec(ElseFlag);
      if ElseFlag=0 then
        dec(IfNest);
    end;
    ExecCmnd := Err;
    exit;
  end;

  Err := 0;
  if GetReservedWord(WId) then
    case WId of
      RsvBeep:       Err := TTLBeep;
      RsvBPlusRecv:  Err := TTLBPlusRecv;
      RsvBPlusSend:  Err := TTLBPlusSend;
      RsvCall:       Err := TTLCall;
      RsvChangeDir:  Err := TTLChangeDir;
      RsvCloseTT:    Err := TTLCloseTT;
      RsvConnect:    Err := TTLConnect;
      RsvElse:       Err := TTLElse;
      RsvElseIf:     Err := TTLElseIf;
      RsvEnd:        Err := TTLEnd;
      RsvEndIf:      Err := TTLEndIf;
      RsvEndWhile:   Err := TTLEndWhile;
      RsvExec:       Err := TTLExec;
      RsvExecCmnd:   Err := TTLExecCmnd;
      RsvExit:       Err := TTLExit;
      RsvFileClose:  Err := TTLFileClose;
      RsvFileConcat: Err := TTLFileConcat;
      RsvFileCopy:   Err := TTLFileCopy;
      RsvFileCreate: Err := TTLFileCreate;
      RsvFileDelete: Err := TTLFileDelete;
      RsvFileOpen:   Err := TTLFileOpen;
      RsvFileReadln: Err := TTLFileReadln;
      RsvFileRename: Err := TTLFileRename;
      RsvFileSearch: Err := TTLFileSearch;
      RsvFileSeek:   Err := TTLFileSeek;
      RsvFileStrSeek: Err := TTLFileStrSeek;
      RsvFileWrite:  Err := TTLFileWrite;
      RsvFileWriteln: Err := TTLFileWriteln;
      RsvFor:        Err := TTLFor;
      RsvGetDate:    Err := TTLGetDate;
      RsvGetTime:    Err := TTLGetTime;
      RsvGoto:       Err := TTLGoto;
      RsvIf:         Err := TTLIf;
      RsvInclude:    Err := TTLInclude;
      RsvInputBox:   Err := TTLInputBox(FALSE);
      RsvInt2Str:    Err := TTLInt2Str;
      RsvKmtRecv:    Err := TTLKmtRecv;
      RsvKmtSend:    Err := TTLKmtSend;
      RsvLogClose:   Err := TTLLogClose;
      RsvLogOpen:    Err := TTLLogOpen;
      RsvLogPause:   Err := TTLLogPause;
      RsvLogStart:   Err := TTLLogStart;
      RsvLogWrite:   Err := TTLLogWrite;
      RsvMessageBox: Err := TTLMessageBox;
      RsvNext:       Err := TTLNext;
      RsvPasswordBox: Err := TTLInputBox(TRUE);
      RsvPause:      Err := TTLPause;
      RsvQuickVANRecv: Err := TTLQuickVANRecv;
      RsvQuickVANSend: Err := TTLQuickVANSend;
      RsvReturn:     Err := TTLReturn;
      RsvSend:       Err := TTLSend;
      RsvSendFile:   Err := TTLSendFile;
      RsvSendln:     Err := TTLSendln;
      RsvShow:       Err := TTLShow(FALSE);
      RsvShowTT:     Err := TTLShow(TRUE);
      RsvStr2Int:    Err := TTLStr2Int;
      RsvStrCompare: Err := TTLStrCompare;
      RsvStrConcat:  Err := TTLStrConcat;
      RsvStrCopy:    Err := TTLStrCopy;
      RsvStrLen:     Err := TTLStrLen;
      RsvStrScan:    Err := TTLStrScan;
      RsvWait:       Err := TTLWait;
      RsvWaitRecv:   Err := TTLWaitRecv;
      RsvWhile:      Err := TTLWhile;
      RsvXmodemRecv: Err := TTLXmodemRecv;
      RsvXmodemSend: Err := TTLXmodemSend;
      RsvYesNoBox:   Err := TTLYesNoBox;
      RsvZmodemRecv: Err := TTLZmodemRecv;
      RsvZmodemSend: Err := TTLZmodemSend;
    else
      Err := ErrSyntax;
    end
  else if GetIdentifier(Cmnd) then
  begin
    if (GetFirstChar='=') then
    begin
      StrConst := GetString(Str,Err);
      if StrConst then
        ValType := TypString
      else
        if not GetExpression(ValType,Val,Err) then
          Err := ErrSyntax;

      if Err=0 then
      begin
        if CheckVar(Cmnd,VarType,VarId) then
        begin
          if VarType=ValType then
            case ValType of
              TypInteger: IntVal[VarId] := Val;
              TypString:
                if StrConst then
                  SetStrVal(VarId,Str)
                else
                  StrCopy(StrVarPtr(VarId),StrVarPtr(Val));
            end
          else
            Err := ErrTypeMismatch;
        end
        else begin
          case ValType of
            TypInteger: E := NewIntVar(Cmnd,Val);
            TypString:
              if StrConst then
                E := NewStrVar(Cmnd,Str)
              else
                E := NewStrVar(Cmnd,StrVarPtr(Val));
          else
            E := false;
          end;
          if not E then Err := ErrTooManyVar;
        end;
        if (Err=0) and (GetFirstChar<>#0) then
          Err := ErrSyntax;
      end;
    end
    else Err := ErrSyntax;
  end
  else
    Err := ErrSyntax;

  ExecCmnd := Err;
end;

procedure Exec;
var
  Err: word;
begin  
  {ParseAgain is set by 'ExecCmnd'}
  if not ParseAgain and
     not GetNewLine then
  begin
    TTLStatus := IdTTLEnd;
    exit;
  end;
  ParseAgain := FALSE;
  LockVar;
  Err := ExecCmnd;
  if Err>0 then DispErr(Err);
  UnlockVar;
end;

procedure SetInputStr(Str: PChar);
var
  VarType, VarId: word;
begin
  if CheckVar('inputstr',VarType,VarId) and
     (VarType=TypString) then
    SetStrVal(VarId,Str);
end;

procedure SetResult(ResultCode: integer);
var
  VarType, VarId: word;
begin
  if CheckVar('result',VarType,VarId) and
     (VarType=TypInteger) then
    IntVal[VarId] := ResultCode;
end;

procedure SetTimeOut(TimeOut: integer);
var
  VarType, VarId: word;
begin
  if CheckVar('timeout',VarType,VarId) and
     (VarType=TypInteger) then
    IntVal[VarId] := TimeOut;
end;

function CalcTime: longint;
var
  H, M, S, S100: word;
begin
  GetTime(H,M,S,S100);
  CalcTime := H*3600+M*60+S;
end;

begin
  TTLStatus := 0;
end.